import logging
import os
import socket
import sqlite3
from datetime import date, datetime, time, timedelta
from os.path import normpath, split
from uuid import uuid4

import icalendar
from atomicwrites import AtomicWriter
from dateutil.tz import tzlocal

from todoman import exceptions

logger = logging.getLogger(name=__name__)


# Initialize this only once
# We were doing this all over the place (even if unused!), so at least only do
# it once.
LOCAL_TIMEZONE = tzlocal()


class UnsafeOperationException(Exception):
    """
    Raised when attempting to perform an unsafe operation.

    Typical examples of unsafe operations are attempting to save a
    partially-loaded todo.
    """
    pass


class AlreadyExists(Exception):
    """
    Raise when two objects have a same identity.

    This can ocurrs when two lists have the same name, or when two Todos have
    the same path.
    """
    pass


class cached_property:  # noqa
    '''A read-only @property that is only evaluated once. Only usable on class
    instances' methods.
    '''
    def __init__(self, fget, doc=None):
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        self.__doc__ = doc or fget.__doc__
        self.fget = fget

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class Todo:
    """
    Represents a task/todo, and wrapps around icalendar.Todo.

    All text attributes are always treated as text, and "" will be returned if
    they are not defined.
    Date attributes are treated as datetime objects, and None will be returned
    if they are not defined.
    All datetime objects have tzinfo, either the one defined in the file, or
    the local system's one.
    """

    def __init__(self, filename=None, mtime=None, new=False, list=None):
        """
        Creates a new todo using `todo` as a source.

        :param str filename: The name of the file for this todo. Defaults to
        the <uid>.ics
        :param mtime int: The last modified time for the file backing this
        Todo.
        :param bool new: Indicate that a new Todo is being created and should
        be populated with default values.
        :param List list: The list to which this Todo belongs.
        """
        self.list = list
        now = datetime.now(LOCAL_TIMEZONE)
        self.uid = '{}@{}'.format(uuid4().hex, socket.gethostname())
        self.list = list

        if new:
            self.created_at = now
        else:
            self.created_at = None

        # Default values for supported fields
        self.categories = []
        self.completed_at = None
        self.description = ''
        self.dtstamp = now
        self.due = None
        self.id = None
        self.last_modified = None
        self.location = ''
        self.percent_complete = 0
        self.priority = 0
        self.sequence = 0
        self.start = None
        self.status = 'NEEDS-ACTION'
        self.summary = ''

        self.filename = filename or "{}.ics".format(self.uid)

        if os.path.basename(self.filename) != self.filename:
            raise ValueError(
                'Must not be an absolute path: {}' .format(self.filename)
            )
        self.mtime = mtime or datetime.now()

    def clone(self):
        """
        Returns a clone of this todo

        Returns a copy of this todo, which is almost identical, except that is
        has a different UUID and filename.
        """
        todo = Todo(new=True, list=self.list)

        fields = (
            Todo.STRING_FIELDS +
            Todo.INT_FIELDS +
            Todo.LIST_FIELDS +
            Todo.DATETIME_FIELDS
        )
        fields.remove('uid')

        for field in fields:
            setattr(todo, field, getattr(self, field))

        return todo

    STRING_FIELDS = [
        'description',
        'location',
        'status',
        'summary',
        'uid',
    ]
    INT_FIELDS = [
        'percent_complete',
        'priority',
        'sequence',
    ]
    LIST_FIELDS = [
        'categories',
    ]
    DATETIME_FIELDS = [
        'completed_at',
        'created_at',
        'dtstamp',
        'start',
        'due',
        'last_modified',
    ]
    ALL_SUPPORTED_FIELDS = (
        DATETIME_FIELDS +
        INT_FIELDS +
        LIST_FIELDS +
        STRING_FIELDS
    )

    VALID_STATUSES = (
        "CANCELLED",
        "COMPLETED",
        "IN-PROCESS",
        "NEEDS-ACTION",
    )

    def __setattr__(self, name, value):
        # Avoids accidentally setting a field to None when that's not a valid
        # attribute.
        if not value:
            if name in Todo.STRING_FIELDS:
                return object.__setattr__(self, name, '')
            if name in Todo.INT_FIELDS:
                return object.__setattr__(self, name, 0)
            if name in Todo.LIST_FIELDS:
                return object.__setattr__(self, name, [])

        return object.__setattr__(self, name, value)

    @property
    def is_completed(self):
        return (
            bool(self.completed_at) or
            self.status in ('CANCELLED', 'COMPLETED')
        )

    @is_completed.setter
    def is_completed(self, val):
        if val:
            # Don't fiddle with completed_at if this was already completed:
            if not self.is_completed:
                self.completed_at = datetime.now(tz=LOCAL_TIMEZONE)
            self.percent_complete = 100
            self.status = 'COMPLETED'
        else:
            self.completed_at = None
            self.percent_complete = None
            self.status = 'NEEDS-ACTION'

    @cached_property
    def path(self):
        return os.path.join(self.list.path, self.filename)

    def cancel(self):
        self.status = 'CANCELLED'


class VtodoWritter:
    """Writes a Todo as a VTODO file."""

    """Maps Todo field names to VTODO field names"""
    FIELD_MAP = {
        'summary': 'summary',
        'priority': 'priority',
        'sequence': 'sequence',
        'uid': 'uid',
        'categories': 'categories',
        'completed_at': 'completed',
        'description': 'description',
        'dtstamp': 'dtstamp',
        'start': 'dtstart',
        'due': 'due',
        'location': 'location',
        'percent_complete': 'percent-complete',
        'priority': 'priority',
        'status': 'status',
        'created_at': 'created',
        'last_modified': 'last-modified',
    }

    def __init__(self, todo):
        self.todo = todo

    def normalize_datetime(self, dt):
        '''
        Eliminate several differences between dates, times and datetimes which
        are hindering comparison:

        - Convert everything to datetime
        - Add missing timezones
        '''
        if isinstance(dt, date) and not isinstance(dt, datetime):
            dt = datetime(dt.year, dt.month, dt.day)

        if not dt.tzinfo:
            dt = dt.replace(tzinfo=LOCAL_TIMEZONE)

        return dt

    def serialize_field(self, name, value):
        if name in Todo.DATETIME_FIELDS:
            return self.normalize_datetime(value)
        if name in Todo.LIST_FIELDS:
            return ','.join(value)
        if name in Todo.INT_FIELDS:
            return int(value)
        if name in Todo.STRING_FIELDS:
            return value

        raise Exception('Unknown field {} serialized.'.format(name))

    def set_field(self, name, value):
        if name in self.vtodo:
            del(self.vtodo[name])
        if value:
            logger.debug("Setting field %s to %s.", name, value)
            self.vtodo.add(name, value)

    def serialize(self, original=None):
        """Serialize a Todo into a VTODO."""
        if not original:
            original = icalendar.Todo()
        self.vtodo = original

        for source, target in self.FIELD_MAP.items():
            if getattr(self.todo, source):
                self.set_field(
                    target,
                    self.serialize_field(source, getattr(self.todo, source)),
                )

        return self.vtodo

    def _read(self, path):
        with open(path, 'rb') as f:
            cal = f.read()
            cal = icalendar.Calendar.from_ical(cal)
            for component in cal.walk('VTODO'):
                return component

    def write(self):
        if os.path.exists(self.todo.path):
            self._write_existing(self.todo.path)
        else:
            self._write_new(self.todo.path)

        return self.vtodo

    def _write_existing(self, path):
        original = self._read(path)
        vtodo = self.serialize(original)

        with open(path, 'rb') as f:
            cal = icalendar.Calendar.from_ical(f.read())
            for index, component in enumerate(cal.subcomponents):
                if component.get('uid', None) == self.todo.uid:
                    cal.subcomponents[index] = vtodo

        with AtomicWriter(path, overwrite=True).open() as f:
            f.write(cal.to_ical().decode("UTF-8"))

    def _write_new(self, path):
        vtodo = self.serialize()

        c = icalendar.Calendar()
        c.add_component(vtodo)

        with AtomicWriter(path).open() as f:
            c.add('prodid', 'io.barrera.todoman')
            c.add('version', '2.0')
            f.write(c.to_ical().decode("UTF-8"))

        return vtodo


class Cache:
    """
    Caches Todos for faster read and simpler querying interface

    The Cache class persists relevant[1] fields into an SQL database, which is
    only updated if the actual file has been modified. This greatly increases
    load times, but, more importantly, provides a simpler interface for
    filtering/querying/sorting.

    [1]: Relevant fields are those we show when listing todos, or those which
    may be used for filtering/sorting.
    """

    SCHEMA_VERSION = 4

    def __init__(self, path):
        self.cache_path = str(path)
        os.makedirs(os.path.dirname(self.cache_path), exist_ok=True)

        self._conn = sqlite3.connect(self.cache_path)
        self._conn.row_factory = sqlite3.Row
        self._conn.execute("PRAGMA foreign_keys = ON")

        self.create_tables()

    def save_to_disk(self):
        self._conn.commit()

    def is_latest_version(self):
        """Checks if the cache DB schema is the latest version."""
        try:
            return self._conn.execute(
                'SELECT version FROM meta WHERE version = ?',
                (Cache.SCHEMA_VERSION,),
            ).fetchone()
        except sqlite3.OperationalError:
            return False

    def create_tables(self):
        if self.is_latest_version():
            return

        self._conn.executescript('''
            DROP TABLE IF EXISTS lists;
            DROP TABLE IF EXISTS files;
            DROP TABLE IF EXISTS todos;
        ''')

        self._conn.execute(
            'CREATE TABLE IF NOT EXISTS meta ("version" INT)'
        )

        self._conn.execute(
            'INSERT INTO meta (version) VALUES (?)',
            (Cache.SCHEMA_VERSION,),
        )

        self._conn.execute('''
            CREATE TABLE IF NOT EXISTS lists (
                "name" TEXT PRIMARY KEY,
                "path" TEXT,
                "colour" TEXT,
                CONSTRAINT path_unique UNIQUE (path)
            );
        ''')

        self._conn.execute('''
            CREATE TABLE IF NOT EXISTS files (
                "path" TEXT PRIMARY KEY,
                "list_name" TEXT,
                "mtime" INTEGER,

                CONSTRAINT path_unique UNIQUE (path),
                FOREIGN KEY(list_name) REFERENCES lists(name) ON DELETE CASCADE
            );
        ''')

        self._conn.execute('''
            CREATE TABLE IF NOT EXISTS todos (
                "file_path" TEXT,

                "id" INTEGER PRIMARY KEY,
                "uid" TEXT,
                "summary" TEXT,
                "due" INTEGER,
                "start" INTEGER,
                "priority" INTEGER,
                "created_at" INTEGER,
                "completed_at" INTEGER,
                "percent_complete" INTEGER,
                "dtstamp" INTEGER,
                "status" TEXT,
                "description" TEXT,
                "location" TEXT,
                "categories" TEXT,
                "sequence" INTEGER,
                "last_modified" INTEGER,

                FOREIGN KEY(file_path) REFERENCES files(path) ON DELETE CASCADE
            );
        ''')

    def clear(self):
        self._conn.close()
        os.remove(self.cache_path)
        self._conn = None

    def add_list(self, name, path, colour):
        """
        Inserts a new list into the cache.

        Returns the id of the newly inserted list.
        """

        result = self._conn.execute(
            'SELECT name FROM lists WHERE path = ?',
            (path,),
        ).fetchone()

        if result:
            return result['name']

        try:
            self._conn.execute(
                "INSERT INTO lists (name, path, colour) VALUES (?, ?, ?)",
                (name, path, colour,),
            )
        except sqlite3.IntegrityError as e:
            raise AlreadyExists(name) from e

        return self.add_list(name, path, colour)

    def add_file(self, list_name, path, mtime):
        try:
            self._conn.execute('''
                INSERT INTO files (
                    list_name,
                    path,
                    mtime
                ) VALUES (?, ?, ?);
                ''', (
                list_name,
                path,
                mtime,
            ))
        except sqlite3.IntegrityError as e:
            raise AlreadyExists(list_name) from e

    def _serialize_datetime(self, todo, field):
        dt = todo.decoded(field, None)
        if not dt:
            return None

        if isinstance(dt, date) and not isinstance(dt, datetime):
            dt = datetime(dt.year, dt.month, dt.day)
        # XXX: Can we actually read times from files?
        elif isinstance(dt, time):
            dt = datetime.combine(date.today(), dt)

        if not dt.tzinfo:
            dt = dt.replace(tzinfo=LOCAL_TIMEZONE)
        return dt.timestamp()

    def add_vtodo(self, todo, file_path, id=None):
        """
        Adds a todo into the cache.

        :param icalendar.Todo todo: The icalendar component object on which
        """

        sql = '''
            INSERT INTO todos (
                {}
                file_path,
                uid,
                summary,
                due,
                start,
                priority,
                created_at,
                completed_at,
                percent_complete,
                dtstamp,
                status,
                description,
                location,
                categories,
                sequence,
                last_modified
            ) VALUES ({}?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            '''

        due = self._serialize_datetime(todo, 'due')
        start = self._serialize_datetime(todo, 'dtstart')

        if start and due:
            start = None if start >= due else start

        params = (
            file_path,
            todo.get('uid'),
            todo.get('summary'),
            due,
            start,
            todo.get('priority', 0) or None,
            self._serialize_datetime(todo, 'created'),
            self._serialize_datetime(todo, 'completed'),
            todo.get('percent-complete', None),
            self._serialize_datetime(todo, 'dtstamp'),
            todo.get('status', 'NEEDS-ACTION'),
            todo.get('description', None),
            todo.get('location', None),
            todo.get('categories', None),
            todo.get('sequence', 1),
            self._serialize_datetime(todo, 'last-modified'),
        )

        if id:
            params = (id,) + params
            sql = sql.format('id,\n', '?, ')
        else:
            sql = sql.format('', '')

        cursor = self._conn.cursor()
        try:
            cursor.execute(sql, params)
            rv = cursor.lastrowid
        finally:
            cursor.close()

        return rv

    def todos(self, lists=[], priority=None, location='', category='', grep='',
              sort=[], reverse=True, due=None, start=None, startable=False,
              status=['NEEDS-ACTION', 'IN-PROCESS']):
        """
        Returns filtered cached todos, in a specified order.

        If no order is specified, todos are sorted by the following fields::

            completed_at
            -priority
            due
            -created_at

        :param list lists: Only return todos for these lists.
        :param str location: Only return todos with a location containing this
            string.
        :param str category: Only return todos with a category containing this
            string.
        :param str grep: Filter common fields with this substring.
        :param list sort: Order returned todos by these fields. Field names
            with a ``-`` prepended will be used to sort in reverse order.
        :param bool reverse: Reverse the order of the todos after sorting.
        :param int due: Return only todos due within ``due`` hours.
        :param str priority: Only return todos with priority at least as
            high as specified.
        :param tuple(bool, datetime) start: Return only todos before/after
            ``start`` date
        :param list(str) status: Return only todos with any of the given
            statuses.
        :return: A sorted, filtered list of todos.
        :rtype: generator
        """
        extra_where = []
        params = []

        if 'ANY' not in status:
            extra_where.append(
                'AND status IN ({})'.format(', '.join(['?'] * len(status)))
            )
            params.extend(s.upper() for s in status)

        if lists:
            lists = [l.name if isinstance(l, List) else l for l in lists]
            q = ', '.join(['?'] * len(lists))
            extra_where.append('AND files.list_name IN ({})'.format(q))
            params.extend(lists)
        if priority:
            extra_where.append('AND PRIORITY > 0 AND PRIORITY <= ?')
            params.append('{}'.format(priority))
        if location:
            extra_where.append('AND location LIKE ?')
            params.append('%{}%'.format(location))
        if category:
            extra_where.append('AND categories LIKE ?')
            params.append('%{}%'.format(category))
        if grep:
            # # requires sqlite with pcre, which won't be available everywhere:
            # extra_where.append('AND summary REGEXP ?')
            # params.append(grep)
            extra_where.append('AND summary LIKE ?')
            params.append('%{}%'.format(grep))
        if due:
            max_due = (datetime.now() + timedelta(hours=due)).timestamp()
            extra_where.append('AND due IS NOT NULL AND due < ?')
            params.append(max_due)
        if start:
            is_before, dt = start
            dt = dt.timestamp()
            if is_before:
                extra_where.append('AND start <= ?')
                params.append(dt)
            else:
                extra_where.append('AND start >= ?')
                params.append(dt)
        if startable:
            extra_where.append('AND (start IS NULL OR start <= ?)')
            params.append(datetime.now().timestamp())
        if sort:
            order = []
            for s in sort:
                if s.startswith('-'):
                    order.append(' {} ASC'.format(s[1:]))
                else:
                    order.append(' {} DESC'.format(s))
            order = ','.join(order)
        else:
            order = '''
                completed_at DESC,
                priority IS NOT NULL, priority DESC,
                due IS NOT NULL, due DESC,
                created_at ASC
            '''

        if not reverse:
            # Note the change in case to avoid swapping all of them. sqlite
            # doesn't care about casing anyway.
            order = order.replace(' DESC', ' asc').replace(' ASC', ' desc')

        query = '''
              SELECT todos.*, files.list_name, files.path
                FROM todos, files
               WHERE todos.file_path = files.path {}
            ORDER BY {}
        '''.format(' '.join(extra_where), order,)

        logger.debug(query)
        logger.debug(params)

        result = self._conn.execute(query, params)

        seen_paths = set()
        warned_paths = set()

        for row in result:
            todo = self._todo_from_db(row)
            path = row['path']

            if path in seen_paths and path not in warned_paths:
                logger.warning('Todo is in read-only mode because there are '
                               'multiple todos in %s', path)
                warned_paths.add(path)
            seen_paths.add(path)
            yield todo

    def _dt_from_db(self, dt):
        if dt:
            return datetime.fromtimestamp(dt, LOCAL_TIMEZONE)
        return None

    def _todo_from_db(self, row):
        todo = Todo()
        todo.id = row['id']
        todo.uid = row['uid']
        todo.summary = row['summary']
        todo.due = self._dt_from_db(row['due'])
        todo.start = self._dt_from_db(row['start'])
        todo.priority = row['priority']
        todo.created_at = self._dt_from_db(row['created_at'])
        todo.completed_at = self._dt_from_db(row['completed_at'])
        todo.dtstamp = self._dt_from_db(row['dtstamp'])
        todo.percent_complete = row['percent_complete']
        todo.status = row['status']
        todo.description = row['description']
        todo.location = row['location']
        todo.sequence = row['sequence']
        todo.last_modified = row['last_modified']
        todo.list = self.lists_map[row['list_name']]
        todo.filename = os.path.basename(row['path'])
        return todo

    def lists(self):
        result = self._conn.execute("SELECT * FROM lists")
        for row in result:
            yield List(
                name=row['name'],
                path=row['path'],
                colour=row['colour'],
            )

    @cached_property
    def lists_map(self):
        return {l.name: l for l in self.lists()}

    def expire_lists(self, paths):
        results = self._conn.execute("SELECT path, name from lists")
        for result in results:
            if result['path'] not in paths:
                self.delete_list(result['name'])

    def delete_list(self, name):
        self._conn.execute("DELETE FROM lists WHERE lists.name = ?", (name,))

    def todo(self, id, read_only=False):
        # XXX: DON'T USE READ_ONLY
        result = self._conn.execute('''
            SELECT todos.*, files.list_name, files.path
              FROM todos, files
            WHERE files.path = todos.file_path
              AND todos.id = ?
        ''', (id,)
        ).fetchone()

        if not result:
            raise exceptions.NoSuchTodo(id)

        if not read_only:
            count = self._conn.execute('''
                SELECT count(id) AS c
                  FROM files, todos
                 WHERE todos.file_path = files.path
                   AND path=?
            ''', (result['path'],)
            ).fetchone()
            if count['c'] > 1:
                raise exceptions.ReadOnlyTodo(result['path'])

        return self._todo_from_db(result)

    def expire_files(self, paths_to_mtime):
        """Remove stale cache entries based on the given fresh data."""
        result = self._conn.execute("SELECT path, mtime FROM files")
        for row in result:
            path, mtime = row['path'], row['mtime']
            if paths_to_mtime.get(path, None) != mtime:
                self.expire_file(path)

    def expire_file(self, path):
        self._conn.execute("DELETE FROM files WHERE path = ?", (path,))


class List:

    def __init__(self, name, path, colour=None):
        self.path = path
        self.name = name or List.name_for_path(path)
        self.colour = colour or List.colour_for_path(self.path)

    @staticmethod
    def colour_for_path(path):
        try:
            with open(os.path.join(path, 'color')) as f:
                return f.read().strip()
        except (OSError, IOError):
            logger.debug('No colour for list %s', path)

    @staticmethod
    def name_for_path(path):
        try:
            with open(os.path.join(path, 'displayname')) as f:
                return f.read().strip()
        except (OSError, IOError):
            return split(normpath(path))[1]

    @cached_property
    def color_rgb(self):
        color = self.colour
        if not color or not color.startswith('#'):
            return

        r = color[1:3]
        g = color[3:5]
        b = color[5:8]

        if len(r) == len(g) == len(b) == 2:
            return int(r, 16), int(g, 16), int(b, 16)

    @cached_property
    def color_ansi(self):
        rv = self.color_rgb
        if rv:
            return '\33[38;2;{!s};{!s};{!s}m'.format(*rv)

    def __eq__(self, other):
        if isinstance(other, List):
            return self.name == other.name
        return object.__eq__(self, other)

    def __str__(self):
        return self.name


class Database:
    """
    This class is essentially a wrapper around all the lists (which in turn,
    contain all the todos).

    Caching in abstracted inside this class, and is transparent to outside
    classes.
    """

    def __init__(self, paths, cache_path):
        self.cache = Cache(cache_path)
        self.paths = [str(path) for path in paths]
        self.update_cache()

    def update_cache(self):
        self.cache.expire_lists(self.paths)

        paths_to_mtime = {}
        paths_to_list_name = {}

        for path in self.paths:
            list_name = self.cache.add_list(
                List.name_for_path(path),
                path,
                List.colour_for_path(path),
            )
            for entry in os.listdir(path):
                if not entry.endswith('.ics'):
                    continue
                entry_path = os.path.join(path, entry)
                mtime = _getmtime(entry_path)
                paths_to_mtime[entry_path] = mtime
                paths_to_list_name[entry_path] = list_name

        self.cache.expire_files(paths_to_mtime)

        for entry_path, mtime in paths_to_mtime.items():
            list_name = paths_to_list_name[entry_path]

            try:
                self.cache.add_file(list_name, entry_path, mtime)
            except AlreadyExists:
                logger.debug('File already in cache: %s', entry_path)
                continue

            with open(entry_path, 'rb') as f:
                try:
                    cal = f.read()
                    cal = icalendar.Calendar.from_ical(cal)
                    for component in cal.walk('VTODO'):
                        self.cache.add_vtodo(component, entry_path)
                except Exception as e:
                    logger.exception("Failed to read entry %s.", entry_path)

        self.cache.save_to_disk()

    def todos(self, **kwargs):
        return self.cache.todos(**kwargs)

    def todo(self, id, **kwargs):
        return self.cache.todo(id, **kwargs)

    def lists(self):
        return self.cache.lists()

    def move(self, todo, new_list, from_list=None):
        from_list = from_list or todo.list
        orig_path = os.path.join(from_list.path, todo.filename)
        dest_path = os.path.join(new_list.path, todo.filename)

        os.rename(orig_path, dest_path)

    def delete(self, todo):
        path = os.path.join(todo.list.path, todo.filename)
        os.remove(path)

    def flush(self):
        for todo in self.todos(status=['ANY']):
            if todo.is_completed:
                yield todo
                self.delete(todo)

        self.cache.clear()
        self.cache = None

    def save(self, todo):
        todo.sequence += 1
        todo.last_modified = datetime.now(LOCAL_TIMEZONE)

        vtodo = VtodoWritter(todo).write()

        self.cache.expire_file(todo.path)
        mtime = _getmtime(todo.path)

        self.cache.add_file(todo.list.name, todo.path, mtime)
        todo.id = self.cache.add_vtodo(vtodo, todo.path, todo.id)
        self.cache.save_to_disk()


def _getmtime(path):
    stat = os.stat(path)
    return getattr(stat, 'st_mtime_ns', stat.st_mtime)



from unittest import mock

from todoman.widgets import ExtendedEdit, PrioritySelector

# We ignore `size` when testing keypresses, because it's not used anywhere.
# Just pass any number when writing tests, unless we start using the value.

BASE_STRING = 'The lazy fox bla bla\n@ät'


def test_extended_edit_delete_word():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl w')
    assert extended_edit.edit_pos == 4
    assert extended_edit.get_edit_text() == 'The fox bla bla\n@ät'

    extended_edit.set_edit_text('The-lazy-fox-bla-bla\n@ät')
    extended_edit.edit_pos = 8
    extended_edit.keypress(10, 'ctrl w')
    assert extended_edit.edit_pos == 4
    assert extended_edit.get_edit_text() == 'The--fox-bla-bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 6
    extended_edit.keypress(10, 'ctrl w')
    assert extended_edit.edit_pos == 4
    assert extended_edit.get_edit_text() == 'The zy fox bla bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 0
    extended_edit.keypress(10, 'ctrl w')
    assert extended_edit.edit_pos == 0
    assert extended_edit.get_edit_text() == BASE_STRING


def test_extended_edit_delete_sol():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl u')
    assert extended_edit.edit_pos == 0
    assert extended_edit.get_edit_text() == 'fox bla bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 0
    extended_edit.keypress(10, 'ctrl u')
    assert extended_edit.edit_pos == 0
    assert extended_edit.get_edit_text() == BASE_STRING


def test_extended_edit_delete_eol():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl k')
    assert extended_edit.edit_pos == 9
    assert extended_edit.get_edit_text() == 'The lazy \n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 20
    extended_edit.keypress(10, 'ctrl k')
    assert extended_edit.edit_pos == 20
    assert extended_edit.get_edit_text() == BASE_STRING

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 24
    extended_edit.keypress(10, 'ctrl k')
    assert extended_edit.edit_pos == 24
    assert extended_edit.get_edit_text() == BASE_STRING


def test_extended_edit_goto_sol():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl a')
    assert extended_edit.edit_pos == 0
    assert extended_edit.get_edit_text() == BASE_STRING

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 23
    extended_edit.keypress(10, 'ctrl a')
    assert extended_edit.edit_pos == 21
    assert extended_edit.get_edit_text() == BASE_STRING


def test_extended_edit_goto_eol():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl e')
    assert extended_edit.edit_pos == 20
    assert extended_edit.get_edit_text() == BASE_STRING

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 22
    extended_edit.keypress(10, 'ctrl e')
    assert extended_edit.edit_pos == 24
    assert extended_edit.get_edit_text() == BASE_STRING


def test_extended_edit_delete_next_char():
    extended_edit = ExtendedEdit(None)

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 9
    extended_edit.keypress(10, 'ctrl d')
    assert extended_edit.edit_pos == 9
    assert extended_edit.get_edit_text() == 'The lazy ox bla bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 0
    extended_edit.keypress(10, 'ctrl d')
    assert extended_edit.edit_pos == 0
    assert extended_edit.get_edit_text() == 'he lazy fox bla bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 24
    extended_edit.keypress(10, 'ctrl d')
    assert extended_edit.edit_pos == 24
    assert extended_edit.get_edit_text() == 'The lazy fox bla bla\n@ät'

    extended_edit.set_edit_text(BASE_STRING)
    extended_edit.edit_pos = 24
    extended_edit.keypress(10, 'ctrl d')
    assert extended_edit.edit_pos == 24
    assert extended_edit.get_edit_text() == 'The lazy fox bla bla\n@ät'


def test_extended_edit_input():
    """
    Very basic test to make sure we don't break basic editing

    We don't need to do more testing because that's done upstream. We basically
    want to test that we properly forward unhandled keypresses.
    """
    extended_edit = ExtendedEdit(mock.MagicMock())
    extended_edit.keypress((10,), 'h')
    extended_edit.keypress((10,), 'i')

    assert extended_edit.get_edit_text() == 'hi'


def test_extended_edit_editor():
    extended_edit = ExtendedEdit(mock.MagicMock())
    extended_edit.set_edit_text(BASE_STRING)

    with mock.patch('click.edit', return_value='Sheep!') as edit:
        extended_edit.keypress(10, 'ctrl o')

    assert edit.call_count == 1
    assert edit.call_args == mock.call(BASE_STRING)
    assert extended_edit.get_edit_text() == 'Sheep!'


def test_priority_selector(default_formatter):
    selector = PrioritySelector(None, 5, default_formatter.format_priority)

    assert selector.label == 'medium'
    assert selector.priority == 5

    selector.keypress(10, 'right')
    assert selector.label == 'high'
    assert selector.priority == 1

    selector.keypress(10, 'left')
    selector.keypress(10, 'left')
    assert selector.label == 'low'
    assert selector.priority == 9

    selector.keypress(10, 'right')
    assert selector.label == 'medium'
    assert selector.priority == 5

    # Spin the whoel way around:
    for i in PrioritySelector.RANGES:
        selector.keypress(10, 'right')

    assert selector.label == 'medium'
    assert selector.priority == 5

    # Now the other way
    for i in PrioritySelector.RANGES:
        selector.keypress(10, 'left')

    assert selector.label == 'medium'
    assert selector.priority == 5

    # Should do nothing:
    selector.keypress(10, 'd')
    selector.keypress(10, '9')
    assert selector.label == 'medium'
    assert selector.priority == 5



import os
import time
from datetime import datetime
from uuid import uuid4

import pytest
import pytz
from click.testing import CliRunner
from dateutil.tz import tzlocal
from hypothesis import HealthCheck, settings, Verbosity

from todoman import model
from todoman.formatters import DefaultFormatter, HumanizedFormatter


@pytest.fixture
def default_database(tmpdir):
    return model.Database(
        [tmpdir.mkdir('default')],
        tmpdir.mkdir(uuid4().hex).join('cache.sqlite3'),
    )


@pytest.fixture
def config(tmpdir, default_database):
    path = tmpdir.join('config')
    path.write('[main]\n'
               'path = {}/*\n'
               'date_format = %Y-%m-%d\n'
               'time_format = \n'
               'cache_path = {}\n'
               .format(str(tmpdir), str(tmpdir.join('cache.sqlite3'))))
    return path


@pytest.fixture
def runner(config, sleep):

    class SleepyCliRunner(CliRunner):
        """
        Sleeps before invoking to make sure cache entries have expired.
        """
        def invoke(self, *args, **kwargs):
            sleep()
            return super().invoke(*args, **kwargs)

    return SleepyCliRunner(env={
        'TODOMAN_CONFIG': str(config)
    })


@pytest.fixture
def create(tmpdir):
    def inner(name, content, list_name='default'):
        tmpdir.ensure_dir(list_name).join(name).write(
            'BEGIN:VCALENDAR\n'
            'BEGIN:VTODO\n' +
            content +
            'END:VTODO\n'
            'END:VCALENDAR'
        )

    return inner


@pytest.fixture
def now_for_tz():

    def inner(tz='CET'):
        """
        Provides the current time cast to a given timezone.

        This helper should be used in place of datetime.now() when the date
        will be compared to some pre-computed value that assumes a determined
        timezone.
        """
        return datetime.now().replace(tzinfo=tzlocal()) \
            .astimezone(pytz.timezone(tz))

    return inner


@pytest.fixture
def todo_factory(default_database):
    def inner(**attributes):
        todo = model.Todo(new=True)
        todo.list = list(default_database.lists())[0]

        attributes.setdefault('summary', 'YARR!')
        for name, value in attributes.items():
            setattr(todo, name, value)

        default_database.save(todo)

        return todo

    return inner


@pytest.fixture
def default_formatter():
    formatter = DefaultFormatter(tz_override=pytz.timezone('CET'))
    return formatter


@pytest.fixture
def humanized_formatter():
    formatter = HumanizedFormatter(tz_override=pytz.timezone('CET'))
    return formatter


@pytest.fixture(scope='session')
def sleep(tmpdir_factory):
    """
    Sleeps as long as needed for the filesystem's mtime to pick up differences

    Measures how long we need to sleep for the filesystem's mtime precision to
    pick up differences and returns a function that sleeps that amount of time.

    This keeps test fast on systems with high precisions, but makes them pass
    on those that don't (I'm looking at you, macOS).
    """
    tmpfile = tmpdir_factory.mktemp('sleep').join('touch_me')

    def touch_and_mtime():
        tmpfile.open('w').close()
        stat = os.stat(str(tmpfile))
        return getattr(stat, 'st_mtime_ns', stat.st_mtime)

    def inner():
        time.sleep(i)

    i = 0.00001
    while i < 100:
        # Measure three times to avoid things like 12::18:11.9994 [mis]passing
        first = touch_and_mtime()
        time.sleep(i)
        second = touch_and_mtime()
        time.sleep(i)
        third = touch_and_mtime()

        if first != second != third:
            i *= 1.1
            return inner
        i = i * 10

    # This should never happen, but oh, well:
    raise Exception(
        'Filesystem does not seem to save modified times of files. \n'
        'Cannot run tests that depend on this.'
    )


@pytest.fixture
def todos(default_database, sleep):

    def inner(**filters):
        sleep()
        default_database.update_cache()
        return default_database.todos(**filters)

    return inner


settings.register_profile("ci", settings(
    max_examples=1000,
    verbosity=Verbosity.verbose,
    suppress_health_check=[HealthCheck.too_slow]
))
settings.register_profile("deterministic", settings(
    derandomize=True,
))

if os.getenv('DETERMINISTIC_TESTS', 'false').lower() == 'true':
    settings.load_profile("deterministic")
elif os.getenv('CI', 'false').lower() == 'true':
    settings.load_profile("ci")



from datetime import datetime
from unittest import mock

import pytest
import pytz
from freezegun import freeze_time
from urwid import ExitMainLoop

from todoman.interactive import TodoEditor


def test_todo_editor_priority(default_database, todo_factory,
                              default_formatter):
    todo = todo_factory(priority=1)
    lists = list(default_database.lists())

    editor = TodoEditor(todo, lists, default_formatter)
    assert editor._priority.label == 'high'

    editor._priority.keypress(10, 'right')
    with pytest.raises(ExitMainLoop):  # Look at editor._msg_text if this fails
        editor._keypress('ctrl s')

    assert todo.priority is 0


def test_todo_editor_list(default_database, todo_factory, default_formatter,
                          tmpdir):
    tmpdir.mkdir('another_list')

    default_database.paths = [
        str(tmpdir.join('default')),
        str(tmpdir.join('another_list')),
    ]
    default_database.update_cache()

    todo = todo_factory()
    lists = list(default_database.lists())

    editor = TodoEditor(todo, lists, default_formatter)
    default_list = next(filter(
        lambda x: x.label == 'default',
        editor.list_selector
    ))
    another_list = next(filter(
        lambda x: x.label == 'another_list',
        editor.list_selector
    ))

    assert editor.current_list == todo.list
    assert default_list.label == todo.list.name

    another_list.set_state(True)
    editor._save_inner()

    assert editor.current_list == todo.list
    assert another_list.label == todo.list.name


def test_todo_editor_summary(default_database, todo_factory,
                             default_formatter):
    todo = todo_factory()
    lists = list(default_database.lists())

    editor = TodoEditor(todo, lists, default_formatter)
    assert editor._summary.edit_text == 'YARR!'

    editor._summary.edit_text = 'Goodbye'
    with pytest.raises(ExitMainLoop):  # Look at editor._msg_text if this fails
        editor._keypress('ctrl s')

    assert todo.summary == 'Goodbye'


@freeze_time('2017-03-04 14:00:00', tz_offset=4)
def test_todo_editor_due(default_database, todo_factory, default_formatter):
    tz = pytz.timezone('CET')

    todo = todo_factory(due=datetime(2017, 3, 4, 14))
    lists = list(default_database.lists())
    default_formatter.tz = tz

    editor = TodoEditor(todo, lists, default_formatter)
    assert editor._due.edit_text == '2017-03-04 14:00'

    editor._due.edit_text = '2017-03-10 12:00'
    with pytest.raises(ExitMainLoop):  # Look at editor._msg_text if this fails
        editor._keypress('ctrl s')

    assert todo.due == datetime(2017, 3, 10, 12, tzinfo=tz)


def test_toggle_help(default_database, default_formatter, todo_factory):
    todo = todo_factory()
    lists = list(default_database.lists())

    editor = TodoEditor(todo, lists, default_formatter)
    editor._loop = mock.MagicMock()
    assert editor._help_text not in editor.left_column.body.contents

    editor._keypress('f1')
    # Help text is made visible
    assert editor._help_text in editor.left_column.body.contents

    # Called event_loop.draw_screen
    assert editor._loop.draw_screen.call_count == 1
    assert editor._loop.draw_screen.call_args == mock.call()

    editor._keypress('f1')
    # Help text is made visible
    assert editor._help_text not in editor.left_column.body.contents

    # Called event_loop.draw_screen
    assert editor._loop.draw_screen.call_count == 2
    assert editor._loop.draw_screen.call_args == mock.call()


def test_show_save_errors(default_database, default_formatter, todo_factory):
    todo = todo_factory()
    lists = list(default_database.lists())

    editor = TodoEditor(todo, lists, default_formatter)
    # editor._loop = mock.MagicMock()

    editor._due.set_edit_text('not a date')
    editor._keypress('ctrl s')

    assert(
        editor.left_column.body.contents[2].get_text()[0] ==
        'Time description not recognized: not a date'
    )



from unittest.mock import patch

import pytest
from click.testing import CliRunner

from todoman.cli import cli
from todoman.configuration import ConfigurationException, load_config


def test_explicit_nonexistant(runner):
    result = CliRunner().invoke(
        cli,
        env={
            'TODOMAN_CONFIG': '/nonexistant',
        },
        catch_exceptions=True,
    )
    assert result.exception
    assert "Configuration file /nonexistant does not exist" in result.output


def test_xdg_nonexistant(runner):
    with patch('xdg.BaseDirectory.xdg_config_dirs', []):
        result = CliRunner().invoke(
            cli,
            catch_exceptions=True,
        )
        assert result.exception
        assert "No configuration file found" in result.output


def test_xdg_existant(runner, tmpdir, config):
    with tmpdir.mkdir('todoman').join('todoman.conf').open('w') as f:
        f.write(config.open().read())

    with patch('xdg.BaseDirectory.xdg_config_dirs', [str(tmpdir)]):
        result = CliRunner().invoke(
            cli,
            catch_exceptions=True,
        )
        assert not result.exception
        assert not result.output.strip()


def test_sane_config(config, runner, tmpdir):
    config.write(
        '[main]\n'
        'color = auto\n'
        'date_format = %Y-%m-%d\n'
        'path = /\n'
        'cache_path = {}\n'.format(tmpdir.join('cache.sqlite'))
    )
    result = runner.invoke(cli)
    assert not result.exception


def test_invalid_color(config, runner):
    config.write(
        '[main]\n'
        'color = 12\n'
        'path = "/"\n'
    )
    result = runner.invoke(cli, ['list'])
    assert result.exception
    assert 'Error: Bad color setting, the value "12" is unacceptable.' \
        in result.output


def test_invalid_color_arg(config, runner):
    config.write(
        '[main]\n'
        'path = "/"\n'
    )
    result = runner.invoke(cli, ['--color', '12', 'list'])
    assert result.exception
    assert 'Usage:' in result.output


def test_missing_path(config, runner):
    config.write(
        '[main]\n'
        'color = auto\n'
    )
    result = runner.invoke(cli, ['list'])
    assert result.exception
    assert ("Error: path is missing from the ['main'] section of the "
            "configuration file") in result.output


@pytest.mark.xfail(reason="Not implemented")
def test_extra_entry(config, runner):
    config.write(
        '[main]\n'
        'color = auto\n'
        'date_format = %Y-%m-%d\n'
        'path = /\n'
        'blah = false\n'
    )
    result = runner.invoke(cli, ['list'])
    assert result.exception
    assert "Invalid configuration entry" in result.output


@pytest.mark.xfail(reason="Not implemented")
def test_extra_section(config, runner):
    config.write(
        '[main]\n'
        'date_format = %Y-%m-%d\n'
        'path = /\n'
        '[extra]\n'
        'color = auto\n'
    )
    result = runner.invoke(cli, ['list'])
    assert result.exception
    assert "Invalid configuration section" in result.output


def test_missing_cache_dir(config, runner, tmpdir):
    cache_dir = tmpdir.join('does').join('not').join('exist')
    cache_file = cache_dir.join('cache.sqlite')

    path = tmpdir.join('config')
    path.write('cache_path = {}\n'.format(cache_file), 'a')
    path.write('[main]\n'
               'path = {}/*\n'
               'cache_path = {}\n'
               .format(str(tmpdir), cache_file))

    result = runner.invoke(cli)
    assert not result.exception
    assert cache_dir.isdir()
    assert cache_file.isfile()


def test_date_field_in_time_format(config, runner, tmpdir):
    config.write(
        '[main]\n'
        'path = "/"\n'
        'time_format = %Y-%m-%d\n'
    )
    result = runner.invoke(cli)
    assert result.exception
    assert (
        'Found date component in `time_format`, please use `date_format` for '
        'that.' in result.output
    )


def test_date_field_in_time(config, runner, tmpdir):
    config.write(
        '[main]\n'
        'path = "/"\n'
        'date_format = %Y-%d-:%M\n'
    )
    result = runner.invoke(cli)
    assert result.exception
    assert (
        'Found time component in `date_format`, please use `time_format` for '
        'that.' in result.output
    )


def test_colour_validation_auto(config):
    with patch(
        'todoman.configuration.find_config',
        return_value=(str(config)),
    ):
        cfg = load_config()

    assert cfg['main']['color'] == 'auto'


def test_colour_validation_always(config):
    config.write("color = 'always'\n", 'a')
    with patch(
        'todoman.configuration.find_config',
        return_value=(str(config)),
    ):
        cfg = load_config()

    assert cfg['main']['color'] == 'always'


def test_colour_validation_invalid(config):
    config.write("color = 'on_weekends_only'\n", 'a')
    with patch(
        'todoman.configuration.find_config',
        return_value=(str(config)),
    ), pytest.raises(ConfigurationException):
        load_config()



import datetime
import json
from time import mktime

import click
import humanize
import parsedatetime
import pytz
from dateutil.tz import tzlocal
from tabulate import tabulate


class DefaultFormatter:

    def __init__(self, date_format='%Y-%m-%d', time_format='%H:%M',
                 dt_separator=' ', tz_override=None):
        self.date_format = date_format
        self.time_format = time_format
        self.dt_separator = dt_separator
        self.datetime_format = dt_separator.join(filter(bool, (
            date_format, time_format
        )))

        self.tz = tz_override or tzlocal()
        self.now = datetime.datetime.now().replace(tzinfo=self.tz)

        self._parsedatetime_calendar = parsedatetime.Calendar()

    def simple_action(self, action, todo):
        return '{} "{}"'.format(action, todo.summary)

    def compact(self, todo):
        return self.compact_multiple([todo])

    def compact_multiple(self, todos):
        table = []
        for todo in todos:
            completed = "X" if todo.is_completed else " "
            percent = todo.percent_complete or ''
            if percent:
                percent = " ({}%)".format(percent)
            priority = self.format_priority_compact(todo.priority)

            due = self.format_datetime(todo.due)
            if todo.due and todo.due <= self.now and not todo.is_completed:
                due = click.style(due, fg='red')

            table.append([
                todo.id,
                "[{}]".format(completed),
                priority,
                due,
                "{} {}{}".format(
                    todo.summary,
                    self.format_database(todo.list),
                    percent,
                ),
            ])

        return tabulate(table, tablefmt='plain')

    def _columnize(self, label, text):
        rows = []

        lines = text.splitlines()
        rows.append([label, lines[0]])
        for line in lines[1:]:
            rows.append([None, line])

        return rows

    def detailed(self, todo):
        """
        Returns a detailed representation of a task.

        :param Todo todo: The todo component.
        """
        extra_rows = []
        if todo.description:
            extra_rows += self._columnize('Description', todo.description)
        if todo.location:
            extra_rows += self._columnize('Location', todo.location)

        if extra_rows:
            return '{}\n\n{}'.format(
                self.compact(todo), tabulate(extra_rows, tablefmt='plain')
            )
        return self.compact(todo)

    def format_datetime(self, dt):
        if not dt:
            return ''
        elif isinstance(dt, datetime.datetime):
            return dt.strftime(self.datetime_format)
        elif isinstance(dt, datetime.date):
            return dt.strftime(self.date_format)

    def parse_priority(self, priority):
        if priority is None or priority is '':
            return None
        if priority == 'low':
            return 9
        elif priority == 'medium':
            return 5
        elif priority == 'high':
            return 4
        elif priority == 'none':
            return 0
        else:
            raise ValueError('Priority has to be one of low, medium,'
                             ' high or none')

    def format_priority(self, priority):
        if not priority:
            return 'none'
        elif 1 <= priority <= 4:
            return 'high'
        elif priority == 5:
            return 'medium'
        elif 6 <= priority <= 9:
            return 'low'

    def format_priority_compact(self, priority):
        if not priority:
            return ''
        elif 1 <= priority <= 4:
            return "!!!"
        elif priority == 5:
            return "!!"
        elif 6 <= priority <= 9:
            return "!"

    def parse_datetime(self, dt):
        if not dt:
            return None

        rv = self._parse_datetime_naive(dt)
        return rv.replace(tzinfo=self.tz)

    def _parse_datetime_naive(self, dt):
        try:
            return datetime.datetime.strptime(dt, self.datetime_format)
        except ValueError:
            pass

        try:
            return datetime.datetime.strptime(dt, self.date_format)
        except ValueError:
            pass

        try:
            return datetime.datetime.combine(
                self.now.date(),
                datetime.datetime.strptime(dt, self.time_format).time()
            )
        except ValueError:
            pass

        rv, certainty = self._parsedatetime_calendar.parse(dt)
        if not certainty:
            raise ValueError(
                'Time description not recognized: {}' .format(dt)
            )
        return datetime.datetime.fromtimestamp(mktime(rv))

    def format_database(self, database):
        return '{}@{}'.format(database.color_ansi or '',
                              click.style(database.name))


class HumanizedFormatter(DefaultFormatter):

    def format_datetime(self, dt):
        if not dt:
            return ''

        rv = humanize.naturaltime(self.now - dt)
        if ' from now' in rv:
            rv = 'in {}'.format(rv[:-9])

        return rv


class PorcelainFormatter(DefaultFormatter):

    def _todo_as_dict(self, todo):
        return dict(
            completed=todo.is_completed,
            due=self.format_datetime(todo.due),
            id=todo.id,
            list=todo.list.name,
            percent=todo.percent_complete,
            summary=todo.summary,
            priority=todo.priority,
        )

    def compact(self, todo):
        return json.dumps(self._todo_as_dict(todo), indent=4, sort_keys=True)

    def compact_multiple(self, todos):
        data = [self._todo_as_dict(todo) for todo in todos]
        return json.dumps(data, indent=4, sort_keys=True)

    def simple_action(self, action, todo):
        return self.compact(todo)

    def parse_priority(self, priority):
        if priority is None:
            return None
        try:
            if int(priority) in range(0, 10):
                return int(priority)
            else:
                raise ValueError('Priority has to be in the range 0-9')
        except ValueError as e:
            raise click.BadParameter(e)

    def detailed(self, todo):
        return self.compact(todo)

    def format_datetime(self, date):
        if date:
            return int(date.timestamp())
        else:
            return None

    def parse_datetime(self, value):
        if value:
            return datetime.datetime.fromtimestamp(value, tz=pytz.UTC)
        else:
            return None



from datetime import date, datetime, time, timedelta

import pytest
import pytz
from freezegun import freeze_time

from todoman.cli import cli


@pytest.mark.parametrize('interval', [
    (65, 'in a minute'),
    (-10800, '3 hours ago'),
])
@pytest.mark.parametrize('tz', ['CET', 'HST'])
@freeze_time('2017-03-25')
def test_humanized_date(runner, create, interval, now_for_tz, tz):
    seconds, expected = interval
    due = now_for_tz(tz) + timedelta(seconds=seconds)
    create(
        'test.ics',
        'SUMMARY:Hi human!\n'
        'DUE;VALUE=DATE-TIME;TZID={}:{}\n'
        .format(tz, due.strftime('%Y%m%dT%H%M%S'))
    )

    result = runner.invoke(cli, ['--humanize', 'list', '--status', 'ANY'])
    assert not result.exception
    assert expected in result.output


def test_format_priority(default_formatter):
    assert default_formatter.format_priority(None) == 'none'
    assert default_formatter.format_priority(0) == 'none'
    assert default_formatter.format_priority(5) == 'medium'
    for i in range(1, 5):
        assert default_formatter.format_priority(i) == 'high'
    for i in range(6, 10):
        assert default_formatter.format_priority(i) == 'low'


def test_format_priority_compact(default_formatter):
    assert default_formatter.format_priority_compact(None) == ''
    assert default_formatter.format_priority_compact(0) == ''
    assert default_formatter.format_priority_compact(5) == '!!'
    for i in range(1, 5):
        assert default_formatter.format_priority_compact(i) == '!!!'
    for i in range(6, 10):
        assert default_formatter.format_priority_compact(i) == '!'


def test_format_date(default_formatter):
    assert default_formatter.format_datetime(date(2017, 3, 4)) == '2017-03-04'


def test_format_datetime(default_formatter):
    assert default_formatter.format_datetime(datetime(2017, 3, 4, 17, 00)) == \
        '2017-03-04 17:00'


def test_detailed_format(runner, todo_factory):
    todo_factory(
        description='Test detailed formatting\n'
        'This includes multiline descriptions\n'
        'Blah!',
        location='Over the hills, and far away',
    )

    # TODO:use formatter instead of runner?
    result = runner.invoke(cli, ['show', '1'])
    expected = (
        '1  [ ]      YARR! @default\n\n'
        'Description  Test detailed formatting\n'
        '             This includes multiline descriptions\n'
        '             Blah!\n'
        'Location     Over the hills, and far away'
    )
    assert not result.exception
    assert result.output.strip() == expected


def test_parse_time(default_formatter):
    tz = pytz.timezone('CET')
    parsed = default_formatter.parse_datetime('12:00')
    expected = datetime.combine(
        date.today(),
        time(hour=12, minute=0),
    ).replace(tzinfo=tz)
    assert parsed == expected


def test_parse_datetime(default_formatter):
    tz = pytz.timezone('CET')

    parsed = default_formatter.parse_datetime('2017-03-05')
    assert parsed == datetime(2017, 3, 5).replace(tzinfo=tz)

    parsed = default_formatter.parse_datetime('2017-03-05 12:00')
    assert parsed == datetime(2017, 3, 5, 12).replace(tzinfo=tz)

    # Notes. will round to the NEXT matching date, so we need to freeze time
    # for this one:
    with freeze_time('2017-03-04'):
        parsed = default_formatter.parse_datetime(
            'Mon Mar  6 22:50:52 -03 2017'
        )
    assert parsed == datetime(2017, 3, 6, 20, 17).replace(tzinfo=tz)

    assert default_formatter.parse_datetime('') is None

    assert default_formatter.parse_datetime(None) is None


def test_humanized_parse_datetime(humanized_formatter):
    tz = pytz.timezone('CET')

    humanized_formatter.now = datetime(2017, 3, 6, 22, 17).replace(tzinfo=tz)
    dt = datetime(2017, 3, 6, 20, 17).replace(tzinfo=tz)

    assert humanized_formatter.format_datetime(dt) == '2 hours ago'
    assert humanized_formatter.format_datetime(None) == ''


def test_simple_action(default_formatter, todo_factory):
    todo = todo_factory()
    assert default_formatter.simple_action('Delete', todo) == \
        'Delete "YARR!"'


def test_formatting_parsing_consitency(default_formatter):
    tz = pytz.timezone('CET')
    dt = datetime(2017, 3, 8, 21, 6).replace(tzinfo=tz)

    formatted = default_formatter.format_datetime(dt)
    assert default_formatter.parse_datetime(formatted) == dt



import os
import sys
from subprocess import PIPE, Popen

from todoman.cli import cli


def test_main(tmpdir, runner):
    root = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
    env = os.environ.copy()
    env['PYTHONPATH'] = root

    cli_result = runner.invoke(cli, ['--version'])

    pipe = Popen(
        [sys.executable, '-m', 'todoman', '--version'], stdout=PIPE, env=env
    )
    main_output = pipe.communicate()[0]

    assert cli_result.output == main_output.decode()



from datetime import datetime

import pytest
import pytz
from dateutil.tz import tzlocal
from dateutil.tz.tz import tzoffset
from freezegun import freeze_time

from todoman.model import Database, List, Todo


def test_querying(create, tmpdir):
    for list in 'abc':
        for i, location in enumerate('abc'):
            create(
                'test{}.ics'.format(i),
                ('SUMMARY:test_querying\r\n'
                 'LOCATION:{}\r\n').format(location),
                list_name=list
            )

    db = Database(
        [str(tmpdir.ensure_dir(l)) for l in 'abc'],
        str(tmpdir.join('cache'))
    )

    assert len(set(db.todos())) == 9
    assert len(set(db.todos(lists='ab'))) == 6
    assert len(set(db.todos(lists='ab', location='a'))) == 2


def test_retain_tz(tmpdir, create, todos):
    create(
        'ar.ics',
        'SUMMARY:blah.ar\n'
        'DUE;VALUE=DATE-TIME;TZID=HST:20160102T000000\n'
    )
    create(
        'de.ics',
        'SUMMARY:blah.de\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20160102T000000\n'
    )

    todos = list(todos())

    assert len(todos) == 2
    assert todos[0].due == datetime(
        2016, 1, 2, 0, 0, tzinfo=tzoffset(None, -36000)
    )
    assert todos[1].due == datetime(
        2016, 1, 2, 0, 0, tzinfo=tzoffset(None, 3600)
    )


def test_change_paths(tmpdir, create):
    old_todos = set('abcdefghijk')
    for x in old_todos:
        create('{}.ics'.format(x), 'SUMMARY:{}\n'.format(x), x)

    tmpdir.mkdir('3')

    db = Database([tmpdir.join(x) for x in old_todos],
                  tmpdir.join('cache.sqlite'))

    assert {t.summary for t in db.todos()} == old_todos

    db.paths = [str(tmpdir.join('3'))]
    db.update_cache()

    assert len(list(db.lists())) == 1
    assert not list(db.todos())


def test_list_displayname(tmpdir):
    tmpdir.join('default').mkdir()
    with tmpdir.join('default').join('displayname').open('w') as f:
        f.write('personal')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.name == 'personal'
    assert str(list_) == 'personal'


def test_list_colour(tmpdir):
    tmpdir.join('default').mkdir()
    with tmpdir.join('default').join('color').open('w') as f:
        f.write('#8ab6d2')

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour == '#8ab6d2'
    assert list_.color_rgb == (138, 182, 210)
    assert list_.color_ansi == '\x1b[38;2;138;182;210m'


def test_list_no_colour(tmpdir):
    tmpdir.join('default').mkdir()

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite3'))
    list_ = next(db.lists())

    assert list_.colour is None
    assert list_.color_rgb is None
    assert list_.color_ansi is None


def test_database_priority_sorting(create, todos):
    for i in [1, 5, 9, 0]:
        create(
            'test{}.ics'.format(i),
            'PRIORITY:{}\n'.format(i)
        )
    create(
        'test_none.ics'.format(i),
        'SUMMARY:No priority (eg: None)\n'
    )

    todos = list(todos())

    assert todos[0].priority == 0
    assert todos[1].priority == 0
    assert todos[2].priority == 9
    assert todos[3].priority == 5
    assert todos[4].priority == 1


def test_retain_unknown_fields(tmpdir, create, default_database):
    """
    Test that we retain unknown fields after a load/save cycle.
    """
    create(
        'test.ics',
        'UID:AVERYUNIQUEID\n'
        'SUMMARY:RAWR\n'
        'X-RAWR-TYPE:Reptar\n'
    )

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite'))
    todo = db.todo(1, read_only=False)

    todo.description = 'Rawr means "I love you" in dinosaur.'
    default_database.save(todo)

    path = tmpdir.join('default').join('test.ics')
    with path.open() as f:
        vtodo = f.read()
    lines = vtodo.splitlines()

    assert 'SUMMARY:RAWR' in lines
    assert 'DESCRIPTION:Rawr means "I love you" in dinosaur.' in lines
    assert 'X-RAWR-TYPE:Reptar' in lines


def test_todo_setters(todo_factory):
    todo = todo_factory()

    todo.description = 'A tea would be nice, thanks.'
    assert todo.description == 'A tea would be nice, thanks.'

    todo.priority = 7
    assert todo.priority == 7

    now = datetime.now()
    todo.due = now
    assert todo.due == now

    todo.description = None
    assert todo.description == ''

    todo.priority = None
    assert todo.priority == 0

    todo.categories = None
    assert todo.categories == []

    todo.due = None
    assert todo.due is None


@freeze_time('2017-03-19-15')
def test_is_completed():
    completed_at = datetime(2017, 3, 19, 14, tzinfo=pytz.UTC),

    todo = Todo()
    todo.completed_at = completed_at
    todo.percent_complete = 20

    todo.is_completed = True
    assert todo.completed_at == completed_at
    assert todo.percent_complete == 100
    assert todo.status == 'COMPLETED'

    todo.is_completed = False
    assert todo.completed_at is None
    assert todo.percent_complete == 0
    assert todo.status == 'NEEDS-ACTION'

    todo.is_completed = True
    now = datetime(2017, 3, 19, 15, tzinfo=pytz.UTC).astimezone(tzlocal())
    assert todo.completed_at == now
    assert todo.percent_complete == 100
    assert todo.status == 'COMPLETED'


def test_todo_filename_absolute_path():
    Todo(filename='test.ics')
    with pytest.raises(ValueError):
        Todo(filename='/test.ics')


def test_list_equality(tmpdir):
    list1 = List(path=str(tmpdir), name='test list')
    list2 = List(path=str(tmpdir), name='test list')
    list3 = List(path=str(tmpdir), name='yet another test list')

    assert list1 == list2
    assert list1 != list3
    assert list1 != 'test list'


def test_clone():
    now = datetime.now(tz=tzlocal())

    todo = Todo(new=True)
    todo.summary = 'Organize a party'
    todo.location = 'Home'
    todo.due = now
    todo.uid = '123'
    todo.id = '123'
    todo.filename = '123.ics'

    clone = todo.clone()

    assert todo.summary == clone.summary
    assert todo.location == clone.location
    assert todo.due == clone.due
    assert todo.uid != clone.uid
    assert len(clone.uid) > 32
    assert clone.id is None
    assert todo.filename != clone.filename
    assert clone.uid in clone.filename


@freeze_time('2017, 3, 20')
def test_todos_startable(tmpdir, runner, todo_factory, todos):
    todo_factory(summary='started', start=datetime(2017, 3, 15))
    todo_factory(summary='nostart')
    todo_factory(summary='unstarted', start=datetime(2017, 3, 24))

    todos = list(todos(startable=True))

    assert len(todos) == 2
    for todo in todos:
        assert 'unstarted' not in todo.summary


def test_filename_uid_colision(create, default_database, runner, todos):
    create(
        'ABC.ics',
        'SUMMARY:My UID is not ABC\n'
        'UID:NOTABC\n'
    )
    len(list(todos())) == 1

    todo = Todo(new=False)
    todo.uid = 'ABC'
    todo.list = next(default_database.lists())
    default_database.save(todo)

    len(list(todos())) == 2


def test_hide_cancelled(todos, todo_factory):
    todo_factory(status='CANCELLED')

    assert len(list(todos())) == 0
    assert len(list(todos(status='ANY'))) == 1


def test_illegal_start_suppression(create, default_database, todos):
    create(
        'test.ics',
        'SUMMARY:Start doing stuff\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20170331T120000\n'
        'DTSTART;VALUE=DATE-TIME;TZID=CET:20170331T140000\n'
    )
    todo = next(todos())
    assert todo.start is None
    assert todo.due == datetime(2017, 3, 31, 12, tzinfo=tzoffset(None, 7200))


def test_default_status(create, todos):
    create(
        'test.ics',
        'SUMMARY:Finish all these status tests\n'
    )
    todo = next(todos())
    assert todo.status == 'NEEDS-ACTION'



import datetime
import sys
from os.path import isdir
from unittest import mock
from unittest.mock import patch

import click
import hypothesis.strategies as st
import pytest
from dateutil.tz import tzlocal
from freezegun import freeze_time
from hypothesis import given

from todoman.cli import cli, exceptions
from todoman.model import Database, Todo

# TODO: test --grep


def test_list(tmpdir, runner, create):
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert not result.output.strip()

    create(
        'test.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'harhar' in result.output


def test_no_default_list(runner):
    result = runner.invoke(cli, ['new', 'Configure a default list'])

    assert result.exception
    assert ('Error: Invalid value for "--list" / "-l": You must set '
            '"default_list" or use -l.' in result.output)


def test_no_extra_whitespace(tmpdir, runner, create):
    """
    Test that we don't output extra whitespace

    Test that we don't output a lot of extra whitespace when there are no
    tasks, or when there are tasks (eg: both scenarios).

    Note: Other tests should be set up so that comparisons don't care much
    about whitespace, so that if this changes, only this test should fail.
    """
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert result.output == '\n'

    create(
        'test.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert len(result.output.splitlines()) == 1


def test_percent(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
        'PERCENT-COMPLETE:78\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert '78%' in result.output


def test_list_inexistant(tmpdir, runner, create):
    result = runner.invoke(cli, ['list', 'nonexistant'])
    assert result.exception
    assert 'Error: Invalid value for "lists":' in result.output


def test_show_existing(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
        'DESCRIPTION:Lots of text. Yum!\n'
    )
    result = runner.invoke(cli, ['list'])
    result = runner.invoke(cli, ['show', '1'])
    assert not result.exception
    assert 'harhar' in result.output
    assert 'Lots of text. Yum!' in result.output


def test_show_inexistant(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list'])
    result = runner.invoke(cli, ['show', '2'])
    assert result.exit_code == 20
    assert result.output == 'No todo with id 2.\n'


def test_human(runner):
    result = runner.invoke(cli, [
        'new', '-l', 'default', '-d', 'tomorrow', 'hail belzebub'
    ])
    assert not result.exception
    assert 'belzebub' in result.output

    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'belzebub' in result.output


@pytest.mark.xfail(reason='issue#9')
def test_two_events(tmpdir, runner):
    tmpdir.join('default/test.ics').write(
        'BEGIN:VCALENDAR\n'
        'BEGIN:VTODO\n'
        'SUMMARY:task one\n'
        'END:VTODO\n'
        'BEGIN:VTODO\n'
        'SUMMARY:task two\n'
        'END:VTODO\n'
        'END:VCALENDAR'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert len(result.output.splitlines()) == 2
    assert 'task one' in result.output
    assert 'task two' in result.output


def test_default_command(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli)
    assert not result.exception
    assert 'harhar' in result.output


def test_delete(runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    result = runner.invoke(cli, ['delete', '1', '--yes'])
    assert not result.exception
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert not result.output.strip()


def test_delete_prompt(todo_factory, runner, todos):
    todo_factory()

    result = runner.invoke(cli, ['delete', '1'], input='yes')

    assert not result.exception
    assert '[y/N]: yes\nDeleting "YARR!"' in result.output

    assert len(list(todos())) == 0


def test_copy(tmpdir, runner, create):
    tmpdir.mkdir('other_list')
    create(
        'test.ics',
        'SUMMARY:test_copy\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'test_copy' in result.output
    assert 'default' in result.output
    assert 'other_list' not in result.output
    result = runner.invoke(cli, ['copy', '-l', 'other_list', '1'])
    assert not result.exception
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'test_copy' in result.output
    assert 'default' in result.output
    assert 'other_list' in result.output


def test_move(tmpdir, runner, create):
    tmpdir.mkdir('other_list')
    create(
        'test.ics',
        'SUMMARY:test_move\n'
    )
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'test_move' in result.output
    assert 'default' in result.output
    assert 'other_list' not in result.output
    result = runner.invoke(cli, ['move', '-l', 'other_list', '1'])
    assert not result.exception
    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert 'test_move' in result.output
    assert 'default' not in result.output
    assert 'other_list' in result.output


@freeze_time('2017-03-17 20:22:19')
def test_dtstamp(tmpdir, runner, create):
    """Test that we add the DTSTAMP to new entries as per RFC5545."""
    result = runner.invoke(cli, ['new', '-l', 'default', 'test event'])
    assert not result.exception

    db = Database([tmpdir.join('default')],
                  tmpdir.join('/dtstamp_cache'))
    todo = list(db.todos())[0]
    assert todo.dtstamp is not None
    assert todo.dtstamp == datetime.datetime.now(tz=tzlocal())


def test_default_list(tmpdir, runner, create):
    """Test the default_list config parameter"""
    result = runner.invoke(cli, ['new', 'test default list'])
    assert result.exception

    path = tmpdir.join('config')
    path.write('default_list = default\n', 'a')

    result = runner.invoke(cli, ['new', 'test default list'])
    assert not result.exception

    db = Database([tmpdir.join('default')],
                  tmpdir.join('/default_list'))
    todo = list(db.todos())[0]
    assert todo.summary == 'test default list'


@pytest.mark.parametrize(
    'default_due, expected_due_hours', [(None, 24), (1, 1), (0, None)],
    ids=['not specified', 'greater than 0', '0']
)
def test_default_due(
    tmpdir, runner, create, default_due, expected_due_hours
):
    """Test setting the due date using the default_due config parameter"""
    if default_due is not None:
        path = tmpdir.join('config')
        path.write('default_due = {}\n'.format(default_due), 'a')

    runner.invoke(cli, ['new', '-l', 'default', 'aaa'])
    db = Database([tmpdir.join('default')], tmpdir.join('/default_list'))
    todo = list(db.todos())[0]

    if expected_due_hours is None:
        assert todo.due is None
    else:
        assert (todo.due - todo.created_at) == datetime.timedelta(
            hours=expected_due_hours
        )


@freeze_time(datetime.datetime.now())
def test_default_due2(tmpdir, runner, create, todos):
    cfg = tmpdir.join('config')
    cfg.write('default_due = 24\n', 'a')

    r = runner.invoke(cli, ['new', '-ldefault', '-dtomorrow', 'aaa'])
    assert not r.exception
    r = runner.invoke(cli, ['new', '-ldefault', 'bbb'])
    assert not r.exception
    r = runner.invoke(cli, ['new', '-ldefault', '-d', 'one hour', 'ccc'])
    assert not r.exception

    todos = {t.summary: t for t in todos(status='ANY')}
    assert todos['aaa'].due.date() == todos['bbb'].due.date()
    assert todos['ccc'].due == todos['bbb'].due - datetime.timedelta(hours=23)


def test_sorting_fields(tmpdir, runner, default_database):
    tasks = []
    for i in range(1, 10):
        days = datetime.timedelta(days=i)

        todo = Todo(new=True)
        todo.list = next(default_database.lists())
        todo.due = datetime.datetime.now() + days
        todo.created_at = datetime.datetime.now() - days
        todo.summary = 'harhar{}'.format(i)
        tasks.append(todo)

        default_database.save(todo)

    fields = (
        'id',
        'uid',
        'summary',
        'due',
        'priority',
        'created_at',
        'completed_at',
        'dtstamp',
        'status',
        'description',
        'location',
        'categories',
    )

    @given(sort_key=st.lists(
        st.sampled_from(fields + tuple('-' + x for x in fields)),
        unique=True
    ))
    def run_test(sort_key):
        sort_key = ','.join(sort_key)
        result = runner.invoke(cli, ['list', '--sort', sort_key])
        assert not result.exception
        assert result.exit_code == 0
        assert len(result.output.strip().splitlines()) == len(tasks)

    run_test()


def test_sorting_output(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:aaa\n'
        'DUE;VALUE=DATE-TIME;TZID=ART:20160102T000000\n'
    )
    create(
        'test2.ics',
        'SUMMARY:bbb\n'
        'DUE;VALUE=DATE-TIME;TZID=ART:20160101T000000\n'
    )

    examples = [
        ('-summary', ['aaa', 'bbb']),
        ('due', ['aaa', 'bbb'])
    ]

    # Normal sorting, reversed by default
    all_examples = [(['--sort', key], order) for key, order in examples]

    # Testing --reverse, same exact output
    all_examples.extend((['--reverse', '--sort', key], order)
                        for key, order in examples)

    # Testing --no-reverse
    all_examples.extend((['--no-reverse', '--sort', key], reversed(order))
                        for key, order in examples)

    for args, order in all_examples:
        result = runner.invoke(cli, ['list'] + args)
        assert not result.exception
        lines = result.output.splitlines()
        for i, task in enumerate(order):
            assert task in lines[i]


def test_sorting_null_values(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:aaa\n'
        'PRIORITY:9\n'
    )
    create(
        'test2.ics',
        'SUMMARY:bbb\n'
        'DUE;VALUE=DATE-TIME;TZID=ART:20160101T000000\n'
    )

    result = runner.invoke(cli)
    assert not result.exception
    assert 'bbb' in result.output.splitlines()[0]
    assert 'aaa' in result.output.splitlines()[1]


def test_sort_invalid_fields(runner):
    result = runner.invoke(cli, ['list', '--sort', 'hats'])

    assert result.exception
    assert 'Invalid value for "--sort": Unknown field "hats"' in result.output


@pytest.mark.parametrize('hours', [72, -72])
def test_color_due_dates(tmpdir, runner, create, hours):
    due = datetime.datetime.now() + datetime.timedelta(hours=hours)
    create(
        'test.ics',
        'SUMMARY:aaa\n'
        'STATUS:IN-PROCESS\n'
        'DUE;VALUE=DATE-TIME;TZID=ART:{}\n'
        .format(due.strftime('%Y%m%dT%H%M%S'))
    )

    result = runner.invoke(cli, ['--color', 'always'])
    assert not result.exception
    due_str = due.strftime('%Y-%m-%d')
    if hours == 72:
        assert result.output == \
            '1  [ ]    {}  aaa @default\x1b[0m\n'.format(due_str)
    else:
        assert result.output == \
            '1  [ ]    \x1b[31m{}\x1b[0m  aaa @default\x1b[0m\n' \
            .format(due_str)


def test_color_flag(runner, todo_factory):
    todo_factory(due=datetime.datetime(2007, 3, 22))

    result = runner.invoke(cli, ['--color', 'always'], color=True)
    assert(
        result.output.strip() ==
        '1  [ ]    \x1b[31m2007-03-22\x1b[0m  YARR! @default\x1b[0m'
    )
    result = runner.invoke(cli, color=True)
    assert(
        result.output.strip() ==
        '1  [ ]    \x1b[31m2007-03-22\x1b[0m  YARR! @default\x1b[0m'
    )

    result = runner.invoke(cli, ['--color', 'never'], color=True)
    assert(
        result.output.strip() ==
        '1  [ ]    2007-03-22  YARR! @default'
    )


def test_flush(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:aaa\n'
        'STATUS:COMPLETED\n'
    )

    result = runner.invoke(cli, ['list'])
    assert not result.exception

    create(
        'test2.ics',
        'SUMMARY:bbb\n'
    )

    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert '2  [ ]      bbb @default' in result.output

    result = runner.invoke(cli, ['flush'], input='y\n', catch_exceptions=False)
    assert not result.exception

    create(
        'test2.ics',
        'SUMMARY:bbb\n'
    )

    result = runner.invoke(cli, ['list'])
    assert not result.exception
    assert '1  [ ]      bbb @default' in result.output


def test_edit(runner, default_database, todos):
    todo = Todo(new=True)
    todo.list = next(default_database.lists())
    todo.summary = 'Eat paint'
    todo.due = datetime.datetime(2016, 10, 3)
    default_database.save(todo)

    result = runner.invoke(cli, ['edit', '1', '--due', '2017-02-01'])
    assert not result.exception
    assert '2017-02-01' in result.output

    todo = next(todos(status='ANY'))
    assert todo.due == datetime.datetime(2017, 2, 1, tzinfo=tzlocal())
    assert todo.summary == 'Eat paint'


def test_edit_move(runner, todo_factory, default_database, tmpdir, todos):
    """
    Test that editing the list in the UI edits the todo as expected

    The goal of this test is not to test the editor itself, but rather the
    `edit` command and its slightly-complex moving logic.
    """
    tmpdir.mkdir('another_list')

    default_database.paths = [
        str(tmpdir.join('default')),
        str(tmpdir.join('another_list')),
    ]
    default_database.update_cache()

    todo_factory(summary='Eat some headphones')

    lists = list(default_database.lists())
    another_list = next(filter(lambda x: x.name == 'another_list', lists))

    def moving_edit(self):
        self.current_list = another_list
        self._save_inner()

    with patch('todoman.interactive.TodoEditor.edit', moving_edit):
        result = runner.invoke(cli, ['edit', '1'])

    assert not result.exception

    todos = list(todos())
    assert len(todos) == 1
    assert todos[0].list.name == 'another_list'


def test_edit_retains_id(runner, todos, todo_factory):
    """Tests that we retain a todo's ID after editing."""
    original_id = todo_factory().id

    result = runner.invoke(cli, ['edit', '1', '--due', '2017-04-01'])
    assert not result.exception

    todo = next(todos())
    assert todo.due == datetime.datetime(2017, 4, 1, tzinfo=tzlocal())
    assert todo.id == original_id


def test_edit_inexistant(runner):
    """Tests that we show the right output and exit code for inexistant ids."""
    result = runner.invoke(cli, ['edit', '1', '--due', '2017-04-01'])
    assert result.exception
    assert result.exit_code == exceptions.NoSuchTodo.EXIT_CODE
    assert result.output.strip() == 'No todo with id 1.'


def test_empty_list(tmpdir, runner, create):
    for item in tmpdir.listdir():
        if isdir(str(item)):
            item.remove()

    result = runner.invoke(cli)
    expected = ("No lists found matching {}/*, create"
                " a directory for a new list").format(tmpdir)

    assert expected in result.output


def test_show_location(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
        'LOCATION:Boston\n'
    )

    result = runner.invoke(cli, ['show', '1'])
    assert 'Boston' in result.output


def test_location(runner):
    result = runner.invoke(cli, [
        'new', '-l', 'default', '--location', 'Chembur', 'Event Test'
    ])

    assert 'Chembur' in result.output


def test_sort_mixed_timezones(runner, create):
    """
    Test sorting mixed timezones.

    The times on this tests are carefully chosen so that a TZ-naive comparison
    gives the opposite results.
    """
    create(
        'test.ics',
        'SUMMARY:first\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20170304T180000\n'  # 1700 UTC
    )
    create(
        'test2.ics',
        'SUMMARY:second\n'
        'DUE;VALUE=DATE-TIME;TZID=HST:20170304T080000\n'  # 1800 UTC
    )

    result = runner.invoke(cli, ['list', '--status', 'ANY'])
    assert not result.exception
    output = result.output.strip()
    assert len(output.splitlines()) == 2
    assert 'second' in result.output.splitlines()[0]
    assert 'first' in result.output.splitlines()[1]


def test_humanize_interactive(runner):
    result = runner.invoke(cli, ['--humanize', '--porcelain', 'list'])

    assert result.exception
    assert result.output.strip() == \
        "Error: --porcelain and --humanize cannot be used at the same time."


def test_due_bad_date(runner):
    result = runner.invoke(cli, ['new', '--due', 'Not a date', 'Blargh!'])

    assert result.exception
    assert (
        'Error: Invalid value for "--due" / "-d": Time description not '
        'recognized: Not a date' == result.output.strip().splitlines()[-1]
    )


def test_multiple_todos_in_file(runner, create):
    create(
        'test.ics',
        'SUMMARY:a\n'
        'END:VTODO\n'
        'BEGIN:VTODO\n'
        'SUMMARY:b\n'
    )

    for _ in range(2):
        result = runner.invoke(cli, ['list'])
        assert ' a ' in result.output
        assert ' b ' in result.output
        assert 'warning: Todo is in read-only mode' in result.output

    result = runner.invoke(cli, ['done', '1'])
    assert result.exception
    assert 'Todo is in read-only mode because there are multiple todos' \
        in result.output

    result = runner.invoke(cli, ['show', '1'])
    assert not result.exception
    result = runner.invoke(cli, ['show', '2'])
    assert not result.exception


def test_todo_new(runner, default_database):
    # This isn't a very thurough test, but at least catches obvious regressions
    # like startup crashes or typos.

    with patch('urwid.MainLoop'):
        result = runner.invoke(cli, ['new', '-l', 'default'])

    # No SUMMARY error after UI runs
    assert isinstance(result.exception, SystemExit)
    assert result.exception.args == (2,)
    assert 'Error: No SUMMARY specified' in result.output


def test_todo_edit(runner, default_database, todo_factory):
    # This isn't a very thurough test, but at least catches obvious regressions
    # like startup crashes or typos.
    todo_factory()

    with patch('urwid.MainLoop'):
        result = runner.invoke(cli, ['edit', '1'])

    assert not result.exception
    assert 'YARR!' in result.output


@freeze_time('2017, 3, 20')
def test_list_startable(tmpdir, runner, todo_factory):
    todo_factory(summary='started', start=datetime.datetime(2017, 3, 15))
    todo_factory(summary='nostart')
    todo_factory(summary='unstarted', start=datetime.datetime(2017, 3, 24))

    result = runner.invoke(
        cli,
        ['list', '--startable'],
        catch_exceptions=False,
    )

    assert not result.exception
    assert 'started' in result.output
    assert 'nostart' in result.output
    assert 'unstarted' not in result.output

    result = runner.invoke(
        cli,
        ['list'],
        catch_exceptions=False,
    )

    assert not result.exception
    assert 'started' in result.output
    assert 'nostart' in result.output
    assert 'unstarted' in result.output

    path = tmpdir.join('config')
    path.write('startable = yes\n', 'a')

    result = runner.invoke(cli, ['list'], catch_exceptions=False)

    assert not result.exception
    assert 'started' in result.output
    assert 'nostart' in result.output
    assert 'unstarted' not in result.output


def test_bad_start_date(runner):
    result = runner.invoke(cli, ['list', '--start'])
    assert result.exception
    assert (
        result.output.strip() == 'Error: --start option requires 2 arguments'
    )

    result = runner.invoke(cli, ['list', '--start', 'before'])
    assert result.exception
    assert (
        result.output.strip() == 'Error: --start option requires 2 arguments'
    )

    result = runner.invoke(cli, ['list', '--start', 'before', 'not_a_date'])
    assert result.exception
    assert (
        'Invalid value for "--start": Time description not recognized: '
        'not_a_date' in result.output
    )

    result = runner.invoke(cli, ['list', '--start', 'godzilla', '2017-03-22'])
    assert result.exception
    assert ("Format should be '[before|after] [DATE]'" in result.output)


def test_done(runner, todo_factory, todos):
    todo = todo_factory()

    result = runner.invoke(cli, ['done', '1'])
    assert not result.exception

    todo = next(todos(status='ANY'))
    assert todo.percent_complete == 100
    assert todo.is_completed is True

    result = runner.invoke(cli, ['done', '17'])
    assert result.exception
    assert result.output.strip() == 'No todo with id 17.'


def test_cancel(runner, todo_factory, todos):
    todo = todo_factory()

    result = runner.invoke(cli, ['cancel', '1'])
    assert not result.exception

    todo = next(todos(status='ANY'))
    assert todo.status == 'CANCELLED'


def test_id_printed_for_new(runner):
    result = runner.invoke(cli, [
        'new', '-l', 'default', 'show me an id'
    ])
    assert not result.exception
    assert result.output.strip().startswith('1')


def test_repl(runner):
    """Test that repl registers properly."""
    if 'click_repl' not in sys.modules:
        pytest.skip('Optional dependency "click_repl" is not installed')

    result = runner.invoke(cli, ['--help'])

    assert not result.exception
    assert 'repl    Start an interactive shell.' in result.output
    assert 'shell   Start an interactive shell.' in result.output


def test_no_repl(runner):
    """Test that we work fine without click_repl installed."""
    modules = sys.modules
    if 'click_repl' in modules:
        pytest.skip("Test can't be run with click_repl installed")

    result = runner.invoke(cli, ['--help'])

    assert not result.exception
    assert 'repl' not in result.output
    assert 'shell' not in result.output
    assert 'Start an interactive shell.' not in result.output


def test_status_validation():
    from todoman import cli

    @given(statuses=st.lists(
        st.sampled_from(Todo.VALID_STATUSES + ('ANY',)),
        min_size=1,
        max_size=5,
        unique=True
    ))
    def run_test(statuses):
        validated = cli.validate_status(val=','.join(statuses))

        if 'ANY' in statuses:
            assert len(validated) == 4
        else:
            assert len(validated) == len(statuses)

        for status in validated:
            assert status in Todo.VALID_STATUSES

    run_test()


def test_bad_status_validation():
    from todoman import cli
    with pytest.raises(click.BadParameter):
        cli.validate_status(val='INVALID')

    with pytest.raises(click.BadParameter):
        cli.validate_status(val='IN-PROGRESS')


def test_status_filtering(runner, todo_factory):
    todo_factory(summary='one', status='CANCELLED')
    todo_factory(summary='two')

    result = runner.invoke(cli, ['list', '--status', 'cancelled'])
    assert not result.exception
    assert len(result.output.splitlines()) == 1
    assert 'one 'in result.output

    result = runner.invoke(cli, ['list', '--status', 'NEEDS-action'])
    assert not result.exception
    assert len(result.output.splitlines()) == 1
    assert 'two' in result.output


def test_invoke_command(runner, tmpdir):
    path = tmpdir.join('config')
    path.write('default_command = flush\n', 'a')

    flush = mock.MagicMock()
    with patch.dict(cli.commands, values=dict(flush=flush)):
        result = runner.invoke(cli, catch_exceptions=False)

    assert not result.exception
    assert not result.output.strip()
    assert flush.call_count == 1


def test_invoke_invalid_command(runner, tmpdir):
    path = tmpdir.join('config')
    path.write('default_command = DoTheRobot\n', 'a')

    result = runner.invoke(cli, catch_exceptions=False)

    assert result.exception
    assert (
        'Error: Invalid setting for [main][default_command]' in result.output
    )



from datetime import datetime, timedelta

from todoman.cli import cli
from todoman.model import Database, Todo


def test_priority(tmpdir, runner, create):
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert not result.output.strip()

    create(
        'one.ics',
        'SUMMARY:haha\n'
        'PRIORITY:4\n'
    )
    create(
        'two.ics',
        'SUMMARY:hoho\n'
        'PRIORITY:9\n'
    )
    create(
        'three.ics',
        'SUMMARY:hehe\n'
        'PRIORITY:5\n'
    )
    create(
        'four.ics',
        'SUMMARY:huhu\n'
    )

    result_high = runner.invoke(cli, ['list', '--priority=high'])
    assert not result_high.exception
    assert 'haha' in result_high.output
    assert 'hoho' not in result_high.output
    assert 'huhu' not in result_high.output
    assert 'hehe' not in result_high.output

    result_medium = runner.invoke(cli, ['list', '--priority=medium'])
    assert not result_medium.exception
    assert 'haha' in result_medium.output
    assert 'hehe' in result_medium.output
    assert 'hoho' not in result_medium.output
    assert 'huhu' not in result_medium.output

    result_low = runner.invoke(cli, ['list', '--priority=low'])
    assert not result_low.exception
    assert 'haha' in result_low.output
    assert 'hehe' in result_low.output
    assert 'hoho' in result_low.output
    assert 'huhu' not in result_low.output

    result_none = runner.invoke(cli, ['list', '--priority=none'])
    assert not result_none.exception
    assert 'haha' in result_none.output
    assert 'hehe' in result_none.output
    assert 'hoho' in result_none.output
    assert 'huhu' in result_none.output

    result_error = runner.invoke(cli, ['list', '--priority=blah'])
    assert result_error.exception


def test_location(tmpdir, runner, create):
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert not result.output.strip()

    create(
        'one.ics',
        'SUMMARY:haha\n'
        'LOCATION: The Pool\n'
    )
    create(
        'two.ics',
        'SUMMARY:hoho\n'
        'LOCATION: The Dungeon\n'
    )
    create(
        'two.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list', '--location', 'Pool'])
    assert not result.exception
    assert 'haha' in result.output
    assert 'hoho' not in result.output
    assert 'harhar' not in result.output


def test_category(tmpdir, runner, create):
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert not result.output.strip()

    create(
        'one.ics',
        'SUMMARY:haha\n'
        'CATEGORIES:work,trip\n'
    )
    create(
        'two.ics',
        'CATEGORIES:trip\n'
        'SUMMARY:hoho\n'
    )
    create(
        'three.ics',
        'SUMMARY:harhar\n'
    )
    result = runner.invoke(cli, ['list', '--category', 'work'])
    assert not result.exception
    assert 'haha' in result.output
    assert 'hoho' not in result.output
    assert 'harhar' not in result.output


def test_grep(tmpdir, runner, create):
    result = runner.invoke(cli, ['list'], catch_exceptions=False)
    assert not result.exception
    assert not result.output.strip()

    create(
        'one.ics',
        'SUMMARY:fun\n'
        'DESCRIPTION: Have fun!\n',
    )
    create(
        'two.ics',
        'SUMMARY:work\n'
        'DESCRIPTION: The stuff for work\n',
    )
    create(
        'three.ics',
        'SUMMARY:buy sandwiches\n'
        'DESCRIPTION: This is for the Duke\n',
    )
    create(
        'four.ics',
        'SUMMARY:puppies\n'
        'DESCRIPTION: Feed the puppies\n',
    )
    create(
        'five.ics',
        'SUMMARY:research\n'
        'DESCRIPTION: Cure cancer\n',
    )
    create(
        'six.ics',
        'SUMMARY:hoho\n'
    )
    result = runner.invoke(cli, ['list', '--grep', 'fun'])
    assert not result.exception
    assert 'fun' in result.output
    assert 'work' not in result.output
    assert 'sandwiches' not in result.output
    assert 'puppies' not in result.output
    assert 'research' not in result.output
    assert 'hoho' not in result.output


def test_filtering_lists(tmpdir, runner, create):
    tmpdir.mkdir('list_one')
    tmpdir.mkdir('list_two')
    tmpdir.mkdir('list_three')

    runner.invoke(cli, ['new', '-l', 'list_one', 'todo one'])
    runner.invoke(cli, ['new', '-l', 'list_two', 'todo two'])
    runner.invoke(cli, ['new', '-l', 'list_three', 'todo three'])

    result = runner.invoke(cli, ['new', 'list'])
    assert len(result.output.splitlines()) == 3

    result = runner.invoke(cli, ['list', 'list_two'])
    assert not result.exception
    assert len(result.output.splitlines()) == 1
    assert 'todo two' in result.output


def test_due_aware(tmpdir, runner, create, now_for_tz):
    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite'))
    l = next(db.lists())

    for tz in ['CET', 'HST']:
        for i in [1, 23, 25, 48]:
            todo = Todo(new=True)
            todo.due = now_for_tz(tz) + timedelta(hours=i)
            todo.summary = '{}'.format(i)

            todo.list = l
            db.save(todo)

    todos = list(db.todos(due=24))

    assert len(todos) == 4
    assert todos[0].summary == "23"
    assert todos[1].summary == "23"
    assert todos[2].summary == "1"
    assert todos[3].summary == "1"


def test_due_naive(tmpdir, runner, create):
    now = datetime.now()

    for i in [1, 23, 25, 48]:
        due = now + timedelta(hours=i)
        create(
            'test_{}.ics'.format(i),
            'SUMMARY:{}\n'
            'DUE;VALUE=DATE-TIME:{}\n'.format(
                i, due.strftime("%Y%m%dT%H%M%S"),
            )
        )

    db = Database([tmpdir.join('default')], tmpdir.join('cache.sqlite'))
    todos = list(db.todos(due=24))

    assert len(todos) == 2
    assert todos[0].summary == "23"
    assert todos[1].summary == "1"


def test_filtering_start(tmpdir, runner, todo_factory):
    today = datetime.now()
    now = today.strftime("%Y-%m-%d")

    tomorrow = (today + timedelta(days=1)).strftime("%Y-%m-%d")
    yesterday = (today + timedelta(days=-1)).strftime("%Y-%m-%d")

    result = runner.invoke(cli, ['list', '--start', 'before', now])
    assert not result.exception
    assert not result.output.strip()

    result = runner.invoke(cli, ['list', '--start', 'after', now])
    assert not result.exception
    assert not result.output.strip()

    todo_factory(summary='haha', start=today)
    todo_factory(summary='hoho', start=today)
    todo_factory(summary='hihi', start=today - timedelta(days=2))
    todo_factory(summary='huhu')

    result = runner.invoke(cli, ['list', '--start', 'after', yesterday])
    assert not result.exception
    assert 'haha' in result.output
    assert 'hoho' in result.output
    assert 'hihi' not in result.output
    assert 'huhu' not in result.output

    result = runner.invoke(cli, ['list', '--start', 'before', yesterday])
    assert not result.exception
    assert 'haha' not in result.output
    assert 'hoho' not in result.output
    assert 'hihi' in result.output
    assert 'huhu' not in result.output

    result = runner.invoke(cli, ['list', '--start', 'after', tomorrow])
    assert not result.exception
    assert 'haha' not in result.output
    assert 'hoho' not in result.output
    assert 'hihi' not in result.output
    assert 'huhu' not in result.output


def test_statuses(todo_factory, todos):
    cancelled = todo_factory(status='CANCELLED').uid
    completed = todo_factory(status='COMPLETED').uid
    in_process = todo_factory(status='IN-PROCESS').uid
    needs_action = todo_factory(status='NEEDS-ACTION').uid
    no_status = todo_factory(status='NEEDS-ACTION').uid

    all_todos = set(todos(status=['ANY']))
    cancelled_todos = set(todos(status=['CANCELLED']))
    completed_todos = set(todos(status=['COMPLETED']))
    in_process_todos = set(todos(status=['IN-PROCESS']))
    needs_action_todos = set(todos(status=['NEEDS-ACTION']))

    assert {t.uid for t in all_todos} == {
        cancelled,
        completed,
        in_process,
        needs_action,
        no_status
    }
    assert {t.uid for t in cancelled_todos} == {cancelled}
    assert {t.uid for t in completed_todos} == {completed}
    assert {t.uid for t in in_process_todos} == {in_process}
    assert {t.uid for t in needs_action_todos} == {needs_action, no_status}



import os
from os.path import exists, join

import xdg.BaseDirectory
from configobj import ConfigObj, flatten_errors
from validate import Validator, VdtValueError

from todoman import __documentation__


class ConfigurationException(Exception):
    def __init__(self, msg):
        super().__init__((
            '{}\nFor details on the configuration format and a sample file, '
            'see\n{}configure.html'
        ).format(msg, __documentation__))


def expand_path(path):
    """expands `~` as well as variable names"""
    return os.path.expanduser(os.path.expandvars(path))


def validate_cache_path(path):
    if path:
        return expand_path(path)
    else:
        return os.path.join(
            xdg.BaseDirectory.xdg_cache_home,
            'todoman/cache.sqlite3',
        )


def validate_date_format(fmt):
    if any(x in fmt for x in ('%H', '%M', '%S', '%X')):
        raise ConfigurationException(
            'Found time component in `date_format`, please use `time_format` '
            'for that.'
        )
    return fmt


def validate_time_format(fmt):
    if any(x in fmt for x in ('%Y', '%y', '%m', '%d', '%x')):
        raise ConfigurationException(
            'Found date component in `time_format`, please use `date_format` '
            'for that.'
        )
    return fmt


def find_config():
    custom_path = os.environ.get('TODOMAN_CONFIG')
    if custom_path:
        if not exists(custom_path):
            raise ConfigurationException(
                "Configuration file {} does not exist".format(custom_path)
            )
        return custom_path

    for d in xdg.BaseDirectory.xdg_config_dirs:
        path = join(d, 'todoman', 'todoman.conf')
        if exists(path):
            return path

    raise ConfigurationException(
        "No configuration file found.\n\n"
    )


def load_config():
    path = find_config()
    specpath = os.path.join(os.path.dirname(__file__), 'confspec.ini')
    validator = Validator({
        'expand_path': expand_path,
        'cache_path': validate_cache_path,
        'date_format': validate_date_format,
        'time_format': validate_time_format,
    })

    config = ConfigObj(path, configspec=specpath, file_error=True)
    validation = config.validate(validator, preserve_errors=True)

    for section, key, error in flatten_errors(config, validation):
        if not error:
            raise ConfigurationException(
                ('{} is missing from the {} section of the configuration ' +
                 'file').format(key, section)
            )
        if isinstance(error, VdtValueError):
            raise ConfigurationException(
                'Bad {} setting, {}'.format(key, error.args[0])
            )

    return config



class TodomanException(Exception):
    """
    Base class for all our exceptions.

    Should not be raised directly.
    """
    pass


class NoSuchTodo(TodomanException):
    EXIT_CODE = 20

    def __str__(self):
        return 'No todo with id {}.'.format(self.args[0])


class ReadOnlyTodo(TodomanException):
    EXIT_CODE = 21

    def __str__(self):
        return (
            'Todo is in read-only mode because there are multiple todos in {}.'
            .format(self.args[0])
        )


class NoListsFound(TodomanException):
    EXIT_CODE = 22

    def __str__(self):
        return (
            'No lists found matching {}, create a directory for a new list.'
            .format(self.args[0])
        )



#!/usr/bin/env python3

from setuptools import setup

setup(
    name='todoman',
    description='A simple CalDav-based todo manager.',
    author='Hugo Osvaldo Barrera',
    author_email='hugo@barrera.io',
    url='https://github.com/pimutils/todoman',
    license='MIT',
    packages=['todoman'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'todo = todoman.cli:cli',
        ]
    },
    install_requires=[
        open('requirements.txt').readlines()
    ],
    long_description=open('README.rst').read(),
    use_scm_version={
        'version_scheme': 'post-release',
        'write_to': 'todoman/version.py',
    },
    setup_requires=['setuptools_scm != 1.12.0', 'pytest-runner'],
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Console :: Curses',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Office/Business :: Scheduling',
        'Topic :: Utilities',
    ]
)



from datetime import datetime

import pytest
import pytz
from dateutil.tz import tzlocal
from freezegun import freeze_time

from todoman.model import Todo, VtodoWritter


def test_serialize_created_at(todo_factory):
    now = datetime.now(tz=pytz.UTC)
    todo = todo_factory(created_at=now)
    vtodo = VtodoWritter(todo).serialize()

    assert vtodo.get('created') is not None


def test_serialize_dtstart(todo_factory):
    now = datetime.now(tz=pytz.UTC)
    todo = todo_factory(start=now)
    vtodo = VtodoWritter(todo).serialize()

    assert vtodo.get('dtstart') is not None


def test_serializer_raises(todo_factory):
    todo = todo_factory()
    writter = VtodoWritter(todo)

    with pytest.raises(Exception):
        writter.serialize_field('nonexistant', 7)


def test_supported_fields_are_serializeable():
    supported_fields = set(Todo.ALL_SUPPORTED_FIELDS)
    serialized_fields = set(VtodoWritter.FIELD_MAP.keys())

    assert supported_fields == serialized_fields


def test_vtodo_serialization(todo_factory):
    """Test VTODO serialization: one field of each type."""
    description = 'A tea would be nice, thanks.'
    todo = todo_factory(
        categories=['tea', 'drinking', 'hot'],
        description=description,
        due=datetime(3000, 3, 21),
        priority=7,
        status='IN-PROCESS',
        summary='Some tea',
    )
    writer = VtodoWritter(todo)
    vtodo = writer.serialize()

    assert str(vtodo.get('categories')) == 'tea,drinking,hot'
    assert str(vtodo.get('description')) == description
    assert vtodo.get('priority') == 7
    assert vtodo.decoded('due') == datetime(3000, 3, 21, tzinfo=tzlocal())
    assert str(vtodo.get('status')) == 'IN-PROCESS'


@freeze_time('2017-04-04 20:11:57')
def test_update_last_modified(todo_factory, todos, tmpdir):
    todo = todo_factory()
    assert todo.last_modified == datetime.now(tzlocal())


def test_sequence_increment(default_database, todo_factory, todos):
    todo = todo_factory()
    assert todo.sequence == 1

    default_database.save(todo)
    assert todo.sequence == 2

    # Relaod (and check the caching flow for the sequence)
    todo = next(todos())
    assert todo.sequence == 2



import json
from datetime import datetime

import pytz

from todoman.cli import cli
from todoman.formatters import PorcelainFormatter


def test_list_all(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:Do stuff\n'
        'STATUS:COMPLETED\n'
        'DUE;VALUE=DATE-TIME;TZID=CET:20160102T000000\n'
        'PERCENT-COMPLETE:26\n'
    )
    result = runner.invoke(cli, ['--porcelain', 'list', '--status', 'ANY'])

    expected = [{
        'completed': True,
        'due': 1451689200,
        'id': 1,
        'list': 'default',
        'percent': 26,
        'priority': 0,
        'summary': 'Do stuff',
    }]

    assert not result.exception
    assert result.output.strip() == json.dumps(expected, indent=4,
                                               sort_keys=True)


def test_list_nodue(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:Do stuff\n'
        'PERCENT-COMPLETE:12\n'
        'PRIORITY:4\n'
    )
    result = runner.invoke(cli, ['--porcelain', 'list'])

    expected = [{
        'completed': False,
        'due': None,
        'id': 1,
        'list': 'default',
        'percent': 12,
        'priority': 4,
        'summary': 'Do stuff',
    }]

    assert not result.exception
    assert result.output.strip() == json.dumps(expected, indent=4,
                                               sort_keys=True)


def test_list_priority(tmpdir, runner, create):
    result = runner.invoke(cli, ['--porcelain', 'list'],
                           catch_exceptions=False)
    assert not result.exception
    assert result.output.strip() == '[]'
    create(
        'one.ics',
        'SUMMARY:haha\n'
        'PRIORITY:4\n'
    )
    create(
        'two.ics',
        'SUMMARY:hoho\n'
        'PRIORITY:9\n'
    )
    create(
        'three.ics',
        'SUMMARY:hehe\n'
        'PRIORITY:5\n'
    )
    create(
        'four.ics',
        'SUMMARY:huhu\n'
    )

    result_high = runner.invoke(cli, ['--porcelain', 'list',
                                '--priority=4'])
    assert not result_high.exception
    assert 'haha' in result_high.output
    assert 'hoho' not in result_high.output
    assert 'huhu' not in result_high.output
    assert 'hehe' not in result_high.output

    result_medium = runner.invoke(cli, ['--porcelain', 'list',
                                  '--priority=5'])
    assert not result_medium.exception
    assert 'haha' in result_medium.output
    assert 'hehe' in result_medium.output
    assert 'hoho' not in result_medium.output
    assert 'huhu' not in result_medium.output

    result_low = runner.invoke(cli, ['--porcelain', 'list',
                               '--priority=9'])
    assert not result_low.exception
    assert 'haha' in result_low.output
    assert 'hehe' in result_low.output
    assert 'hoho' in result_low.output
    assert 'huhu' not in result_low.output

    result_none = runner.invoke(cli, ['--porcelain', 'list',
                                '--priority=0'])
    assert not result_none.exception
    assert 'haha' in result_none.output
    assert 'hehe' in result_none.output
    assert 'hoho' in result_none.output
    assert 'huhu' in result_none.output

    result_error = runner.invoke(cli, ['--porcelain', 'list',
                                 '--priority=18'])
    assert result_error.exception


def test_show(tmpdir, runner, create):
    create(
        'test.ics',
        'SUMMARY:harhar\n'
        'DESCRIPTION:Lots of text. Yum!\n'
        'PRIORITY:5\n'
    )
    result = runner.invoke(cli, ['--porcelain', 'show', '1'])

    expected = {
        'completed': False,
        'due': None,
        'id': 1,
        'list': 'default',
        'percent': 0,
        'priority': 5,
        'summary': 'harhar',
    }

    assert not result.exception
    assert result.output.strip() == json.dumps(expected, indent=4,
                                               sort_keys=True)


def test_simple_action(todo_factory):
    formatter = PorcelainFormatter()
    todo = todo_factory(id=7, location='Downtown')

    expected = {
        "completed": False,
        "due": None,
        "id": 7,
        "list": "default",
        "percent": 0,
        "priority": 0,
        "summary": "YARR!"
    }

    assert formatter.simple_action('Delete', todo) == \
        json.dumps(expected, indent=4, sort_keys=True)


def test_format_datetime():
    formatter = PorcelainFormatter()

    dt = datetime(2017, 3, 8, 0, 0, 17, 457955, tzinfo=pytz.UTC)
    t = 1488931217

    assert formatter.format_datetime(dt) == t


def test_parse_datetime():
    formatter = PorcelainFormatter()

    expected = datetime(2017, 3, 6, 23, 22, 21, 610429, tzinfo=pytz.UTC)
    assert formatter.parse_datetime(1488842541.610429) == expected

    assert formatter.parse_datetime(None) is None
    assert formatter.parse_datetime('') is None


def test_formatting_parsing_consitency():
    tz = pytz.timezone('CET')
    dt = datetime(2017, 3, 8, 21, 6, 19).replace(tzinfo=tz)

    formatter = PorcelainFormatter(tz_override=tz)
    assert formatter.parse_datetime(formatter.format_datetime(dt)) == dt
